program orbits ( input,output )

; const
    maxd	= 500       (* changed by bert, 14-9-93, was 200 *)
  ; maxorbits	= 10000
  ; maxparams	= 10000
  ; tab		= '	'

; type
    sign	= -1..1
  ; vector3	= array [1..3] of integer
  ; sgdvector3	= record x : vector3 ; sg : sign end

; var
    d,orbits,params	: integer
  ; gcdwithd,invmodd	: array [1..maxd] of integer
  ; orbit		: array [1..maxorbits] of
			  record
			    x : vector3
			  ; orbref,parref : integer
			  end
  ; param		: array [1..maxparams] of integer

#include "common.p"

(**********************************************************************)
(*                                                                    *)
(*                  making a list of representatives                  *)
(*                                                                    *)
(**********************************************************************)

; procedure addifok ( v : vector3 )
  (* Add the vector to the table orbit *)
  (* if it is well-defined and minimal. *)
  ; var
      plusv,min : sgdvector3
    ; opp : boolean
  ; begin
      if coprimewithd(v[1],v[2],v[3]) then (* it is well-defined *)
      begin
        plus(v,plusv)
      ; acts4xmul(plusv,min,opp)
      ; if not prec(min.x,v) then
        begin
          orbits := orbits+1
        ; with orbit[orbits] do
          begin
            x := v
          ; if opp
            then orbref := 0
            else orbref := orbits
          end
        end
      end
    end (* addifok *)

; procedure makeorbitlist
  (* Make a list of orbits, ordered according to the their *)
  (* minimal representatives (lexicographically ordered).  *)
  ; var
      x1,x2,x3 : integer
    ; v : vector3
  ; begin
      orbits := 0
    ; for x1 := 1 to d do
      if d mod x1=0 then
      for x2 := x1 to d do
      for x3 := x2 to d do
      begin
        makevec(x1,x2,x3,v)
      ; addifok(v)
      end
    end (* makeorbitlist *)

(**********************************************************************)
(*                                                                    *)
(*                                                                    *)
(*                                                                    *)
(**********************************************************************)

; function ultorbrefofnum ( num : integer ) : integer
  (* The reference to an orbit ultimately   *)
  (* obtained by following orbref in orbit. *)
  ; var
      stop : boolean
  ; begin
      stop := false
    ; repeat
        if orbit[abs(num)].orbref=0
        then
        begin
          num := 0
        ; stop := true
        end
        else
        if abs(orbit[abs(num)].orbref)=abs(num)
        then stop := true
        else num := sgn(num)*orbit[abs(num)].orbref
      until stop
    ; ultorbrefofnum := num
    end (* ultorbrefofnum *)

; function ultorbrefofvec ( v : vector3 ) : integer
  (* The reference to a minimal orbit representative ultimately
     obtained by following orbref
     in orbit, starting at the minimal representative. *)
  ; var
      plusv,min : sgdvector3
    ; opp : boolean
  ; begin
      plus(v,plusv)
    ; acts4xmul(plusv,min,opp)
    ; ultorbrefofvec := min.sg*ultorbrefofnum(orbitnum(min.x))
    end (* ultorbrefofvec *)

(**********************************************************************)
(*                                                                    *)
(*                     making a list of variables                     *)
(*                                                                    *)
(**********************************************************************)

; procedure remakeorbref
  (* Remake the orbref field of orbit. *)
  ; var
      i : integer
  ; begin
      for i := 1 to orbits do
      orbit[i].orbref := ultorbrefofnum(i)
    end (* remakeorbref *)

; procedure equalize ( v1,v2 : vector3 )
  (* Use the fact that the function values of v1 and v2 are equal. *)
  ; var
      or1,or2 : integer
  ; begin
      or1 := ultorbrefofvec(v1)
    ; or2 := ultorbrefofvec(v2)
    ; if abs(or2)<abs(or1) then swapints(or1,or2)
    ; if or2<>0 then (* else there is no information *)
      begin
        if abs(or2)>abs(or1)
        then orbit[abs(or2)].orbref := sgn(or2)*or1
        else (* abs(or1)=abs(or2) *)
        if or1<>or2 (* else there is no information *)
        then orbit[abs(or2)].orbref := 0
      end
    end (* equalize *)

; procedure usevanishing
  (* Use those cases of                                *)
  (* f(x,y,z) + f(-y,x-y,z) + f(y-x,-x,z) = 0          *)
  (* where (at least) one of the three terms vanishes, *)
  (* in order to reduce the number of variables.       *)
  (* If f(x,y,z) vanishes, then the above equation is  *)
  (* equivalent to f(x-y,x,z) = f(x-y,y,z).            *)
  ; var
      i : integer
    ; vanold,vannew : array [1..maxorbits] of boolean
    ; newvan : boolean
    ; v1,v2 : vector3
  ; begin
      for i := 1 to orbits do vanold[i] := false
    ; repeat
        newvan := false
      ; for i := 1 to orbits do
        begin
          vannew[i] := (orbit[i].orbref=0) and not vanold[i]
        ; newvan := newvan or vannew[i]
        end
      ; if newvan then
        begin
          for i := 1 to orbits do
          if vannew[i] then
          with orbit[i] do
          begin
            makevecmodd(x[1]-x[2],x[1],x[3],v1)
          ; makevecmodd(x[1]-x[2],x[2],x[3],v2)
          ; equalize(v1,v2)
          ; makevecmodd(x[1]+x[2],x[1],x[3],v1)
          ; makevecmodd(x[1]+x[2],x[2],x[3],v2)
          ; equalize(v1,v2)
          ; makevecmodd(x[2]-x[3],x[2],x[1],v1)
          ; makevecmodd(x[2]-x[3],x[3],x[1],v2)
          ; equalize(v1,v2)
          ; makevecmodd(x[2]+x[3],x[2],x[1],v1)
          ; makevecmodd(x[2]+x[3],x[3],x[1],v2)
          ; equalize(v1,v2)
          ; makevecmodd(x[3]-x[1],x[3],x[2],v1)
          ; makevecmodd(x[3]-x[1],x[1],x[2],v2)
          ; equalize(v1,v2)
          ; makevecmodd(x[3]+x[1],x[3],x[2],v1)
          ; makevecmodd(x[3]+x[1],x[1],x[2],v2)
          ; equalize(v1,v2)
          ; vanold[i] := true
          end
        ; remakeorbref
        end
      until not newvan
    end (* usevanishing *)

(**********************************************************************)
(*                                                                    *)
(*                           output routine                           *)
(*                                                                    *)
(**********************************************************************)

; procedure writeorbits
  ; var
      i : integer
  ; begin
      writeln(d:1)
    ; writeln(orbits:1)
    ; for i := 1 to orbits do
      with orbit[i] do
      writeln(x[1]:1,tab,x[2]:1,tab,x[3]:1,tab,orbref:1)
    end (* writeorbits *)

; begin
    readln(d)
  ; make2tables
  ; makeorbitlist
  ; usevanishing
  ; writeorbits
  end.
