program solcheck ( input,output )
(* Checks that the guess produced by the LLL algorithm is actual solution. 
*)
; const
    maxd	= 500
  ; maxorbits	= 5000
  ; maxparams	= 5000
  ; maxrows	= 30000
  ; maxdim	= 500
  ; tab		= '	'
  ; bigprime	= 32503

; type
    sign	= -1..1
  ; vector3	= array [1..3] of integer
  ; sgdvector3	= record x : vector3 ; sg : sign end
  ; pointer	= ^node
  ; node	= record par,val : integer ; next : pointer end
  ; colvector	= array [1..maxrows] of integer

; var
    d,orbits,params,rows,diag,dim : integer
  ; gcdwithd,invmodd	: array [1..maxd] of integer
  ; orbit		: array [1..maxorbits] of
			  record
			    x : vector3
			  ; orbref,parref : integer
			  end
  ; param		: array [1..maxparams] of integer
  ; rowptr		: array [1..maxrows] of pointer
  ; parofcol,colofpar	: array [1..maxparams] of integer
  ; thecol		: colvector
  ; badness		: array [1..maxrows] of real (* not useful *)
  ; soln		: array [1..maxdim,1..maxparams] of integer
  ; freepar		: array [1..maxdim] of integer

#include "common.p"

(**********************************************************************)
(*                                                                    *)
(*            elementary operations involving linked lists            *)
(*                                                                    *)
(**********************************************************************)

; procedure insert ( var head,prevptr,insptr,nextptr : pointer )
  (* Insert a new node into a linked list. *)
  ; begin
      new(insptr)
    ; if prevptr=nil
      then head := insptr
      else prevptr^.next := insptr
    ; insptr^.next := nextptr
    end (* insert *)

; procedure delete ( var head,prevptr,delptr : pointer )
  (* Delete a node from a linked list. *)
  ; var
      nextptr : pointer
  ; begin
      nextptr := delptr^.next
    ; dispose(delptr)
    ; if prevptr=nil
      then head := nextptr
      else prevptr^.next := nextptr
    end (* delete *)

; procedure advanceto ( var prevptr,ptr : pointer ; num : integer
                      ; var found : boolean )
  (* Advance the pointers ptr and prevptr until the par
     field of ptr is not smaller than num, or you reach
     the end of the list.  Return in found whether the
     par field of ptr eventually equals num *)
  ; var
      stop : boolean
  ; begin
      stop := false
    ; while (ptr<>nil) and not stop do
        if ptr^.par>=num
        then stop := true
        else
        begin
          prevptr := ptr
        ; ptr := ptr^.next
        end
    ; if ptr=nil
      then found := false
      else found := (ptr^.par=num)
    end (* advanceto *)

(**********************************************************************)
(*                                                                    *)
(*                    making the equations matrix                     *)
(*                                                                    *)
(**********************************************************************)

; procedure initeqnmatrix
  (* Initialize the equations matrix. *)
  ; var
      i : integer
  ; begin
      rows := 0
    ; for i := 1 to params do
      begin
        parofcol[i] := i
      ; colofpar[i] := i
      end
    end (* initeqnmat *)

; procedure newrow
  ; begin
      rows := rows+1
    ; rowptr[rows] := nil
    end (* newrow *)

; procedure remlastrowifzero
  ; begin
      if rowptr[rows]=nil
      then rows := rows-1
    end (* remlastrowifzero *)

; procedure addpartolastrow ( num : integer )
  (* Add the parameter reference num to the
     last row of the equations matrix. *)
  ; var
      ptr,prevptr,newptr : pointer
    ; found : boolean
  ; begin
      if num<>0 then
      begin
        ptr := rowptr[rows]
      ; prevptr := nil
      ; advanceto(prevptr,ptr,abs(num),found)
      ; if found
        then
        begin
          ptr^.val := ptr^.val+sgn(num)
        ; if ptr^.val=0 then delete(rowptr[rows],prevptr,ptr)
        end
        else
        begin
          insert(rowptr[rows],prevptr,newptr,ptr)
        ; newptr^.par := abs(num)
        ; newptr^.val := sgn(num)
        end
      end
    end (* addpartolastrow *)

; procedure condeqn ( v1,v2,v3 : vector3 )
  (* Add the equation f(v1)+f(v2)+f(v3)=0 to the equations matrix,
     provided that the orbits of v2 and v3 vanish or do not precede
     that of v1. *)
  ; var
      on1,on2,on3,pr1,pr2,pr3 : integer
  ; begin
      locate(v1,on1,pr1)
    ; locate(v2,on2,pr2)
    ; locate(v3,on3,pr3)
    ; if ((pr2=0) or (on2>=on1)) and ((pr3=0) or (on3>=on1))
      then
      begin
        newrow
      ; addpartolastrow(pr1)
      ; addpartolastrow(pr2)
      ; addpartolastrow(pr3)
      ; remlastrowifzero
      end
    end (* condeqn *)

; procedure makeeqnmat
  (* Make a matrix with equations.  The equations are sorted by the
     lowest numbered orbit occurring in them which is not known to
     vanish. *)
  ; var
      i : integer
    ; v1,v2,v3 : vector3
  ; begin
      initeqnmatrix
    ; for i := 1 to orbits do
      with orbit[i] do
      if parref<>0 then (* otherwise the function vanishes *)
      begin
        makevecmodd(     x[1],     x[2],x[3],v1)
      ; makevecmodd(     x[2],x[1]-x[2],x[3],v2)
      ; makevecmodd(x[1]-x[2],     x[1],x[3],v3)
      ; condeqn(v1,v2,v3)
      ; makevecmodd(     x[1],     x[2],x[3],v1)
      ; makevecmodd(     x[2],x[1]+x[2],x[3],v2)
      ; makevecmodd(x[1]+x[2],     x[1],x[3],v3)
      ; condeqn(v1,v2,v3)
      ; makevecmodd(     x[2],     x[3],x[1],v1)
      ; makevecmodd(     x[3],x[2]-x[3],x[1],v2)
      ; makevecmodd(x[2]-x[3],     x[2],x[1],v3)
      ; condeqn(v1,v2,v3)
      ; makevecmodd(     x[2],     x[3],x[1],v1)
      ; makevecmodd(     x[3],x[2]+x[3],x[1],v2)
      ; makevecmodd(x[2]+x[3],     x[2],x[1],v3)
      ; condeqn(v1,v2,v3)
      ; makevecmodd(     x[3],     x[1],x[2],v1)
      ; makevecmodd(     x[1],x[3]-x[1],x[2],v2)
      ; makevecmodd(x[3]-x[1],     x[3],x[2],v3)
      ; condeqn(v1,v2,v3)
      ; makevecmodd(     x[3],     x[1],x[2],v1)
      ; makevecmodd(     x[1],x[3]+x[1],x[2],v2)
      ; makevecmodd(x[3]+x[1],     x[3],x[2],v3)
      ; condeqn(v1,v2,v3)
      end
    end (* makeeqnmat *)

(**********************************************************************)
(*                                                                    *)
(*                     Check that the lllresult is solution           *)
(*                                                                    *)
(**********************************************************************)

; function entry ( r,c : integer ) : integer
  (* Give entry (r,c) of the equations matrix. *)
  ; var
      ptr,prevptr : pointer
    ; found : boolean
  ; begin
      ptr := rowptr[r]
    ; prevptr := nil
    ; advanceto(prevptr,ptr,parofcol[c],found)
    ; if found
      then entry := ptr^.val
      else entry := 0
    end (* entry *)


; procedure checksolnspace
  ; var
      s,i,j,sum : integer
  ; begin
    ; for s := 1 to dim do
      begin
      ; for i := 1 to params do thecol[i]:=soln[s,i] 
      ; for i := rows downto 1 do
        begin
          sum := 0
        ; for j := 1 to params do
          if thecol[j]<>0 then sum := sum+entry(i,j)*thecol[j]
        ; if sum<>0 then 
          begin
          writeln('solution incorrect')
          ; soln[maxdim+abs(sum),maxdim+abs(sum)]:=0 (* cause 
                                                        runtime error*)
          end
        end
      end
   end (* calcsolnspace *)

(**********************************************************************)
(*                                                                    *)
(*                       input/output routines                        *)
(*                                                                    *)
(**********************************************************************)
 
; procedure readsubhint
  ; var
      i,s : integer
  ; begin
    ; readln(d)
    ; readln(params)
    ; readln(dim)
    ; for s := 1 to dim do
      begin
        readln(freepar[s])
      end
    ; for i := 1 to dim do
      for s := 1 to params do readln(soln[i,s])
    end

; procedure writesolnspace
  ; var
      i,s : integer
  ; begin
      writeln(d:1)
    ; writeln(params:1)
    ; writeln(dim:1)
    ; for s := 1 to dim do
      begin
        write(freepar[s]:1)
      ; if (s mod 10=0) or (s=dim)
        then writeln
        else write(tab)
      end
    ; for s := 1 to dim do
      for i := 1 to params do
      begin
        write(soln[s,i]:1)
      ; if (i mod 10=0) or (i=params)
        then writeln
        else write(tab)
      end
    end

; begin
    readorbits
  ; readsubhint
  ; make2tables
  ; makeparams
  ; makeeqnmat
  ; checksolnspace 
  ; writesolnspace
  end.
