program solvemod ( input,output )
(* Like solve, but now everything done mod 32503
*)

; const
    maxd	= 500
  ; maxorbits	= 5000
  ; maxparams	= 5000
  ; maxrows	= 30000
  ; maxdim	= 200
  ; tab		= '	'
  ; bigprime	= 32503

; type
    sign	= -1..1
  ; vector3	= array [1..3] of integer
  ; sgdvector3	= record x : vector3 ; sg : sign end
  ; pointer	= ^node
  ; node	= record par,val : integer ; next : pointer end
  ; colvector	= array [1..maxrows] of integer

; var
    d,orbits,params,rows,diag,dim : integer
  ; gcdwithd,invmodd	: array [1..maxd] of integer
  ; orbit		: array [1..maxorbits] of
			  record
			    x : vector3
			  ; orbref,parref : integer
			  end
  ; param		: array [1..maxparams] of integer
  ; rowptr		: array [1..maxrows] of pointer
  ; parofcol,colofpar	: array [1..maxparams] of integer
  ; thecol		: colvector
  ; badness		: array [1..maxrows] of real (* not useful *)
  ; soln		: array [1..maxdim,1..maxparams] of integer
  ; freepar		: array [1..maxdim] of integer

#include "common.p"

(**********************************************************************)
(*                                                                    *)
(*            elementary operations involving linked lists            *)
(*                                                                    *)
(**********************************************************************)

; procedure insert ( var head,prevptr,insptr,nextptr : pointer )
  (* Insert a new node into a linked list. *)
  ; begin
      new(insptr)
    ; if prevptr=nil
      then head := insptr
      else prevptr^.next := insptr
    ; insptr^.next := nextptr
    end (* insert *)

; procedure delete ( var head,prevptr,delptr : pointer )
  (* Delete a node from a linked list. *)
  ; var
      nextptr : pointer
  ; begin
      nextptr := delptr^.next
    ; dispose(delptr)
    ; if prevptr=nil
      then head := nextptr
      else prevptr^.next := nextptr
    end (* delete *)

; procedure advanceto ( var prevptr,ptr : pointer ; num : integer
                      ; var found : boolean )
  (* Advance the pointers ptr and prevptr until the par
     field of ptr is not smaller than num, or you reach
     the end of the list.  Return in found whether the
     par field of ptr eventually equals num *)
  ; var
      stop : boolean
  ; begin
      stop := false
    ; while (ptr<>nil) and not stop do
        if ptr^.par>=num
        then stop := true
        else
        begin
          prevptr := ptr
        ; ptr := ptr^.next
        end
    ; if ptr=nil
      then found := false
      else found := (ptr^.par=num)
    end (* advanceto *)

(**********************************************************************)
(*                                                                    *)
(*                    making the equations matrix                     *)
(*                                                                    *)
(**********************************************************************)

; procedure initeqnmatrix
  (* Initialize the equations matrix. *)
  ; var
      i : integer
  ; begin
      rows := 0
    ; for i := 1 to params do
      begin
        parofcol[i] := i
      ; colofpar[i] := i
      end
    end (* initeqnmat *)

; procedure newrow
  ; begin
      rows := rows+1
    ; rowptr[rows] := nil
    end (* newrow *)

; procedure remlastrowifzero
  ; begin
      if rowptr[rows]=nil
      then rows := rows-1
    end (* remlastrowifzero *)

; procedure addpartolastrow ( num : integer )
  (* Add the parameter reference num to the
     last row of the equations matrix. *)
  ; var
      ptr,prevptr,newptr : pointer
    ; found : boolean
  ; begin
      if num<>0 then
      begin
        ptr := rowptr[rows]
      ; prevptr := nil
      ; advanceto(prevptr,ptr,abs(num),found)
      ; if found
        then
        begin
          ptr^.val := ptr^.val+sgn(num)
        ; if ptr^.val=0 then delete(rowptr[rows],prevptr,ptr)
        end
        else
        begin
          insert(rowptr[rows],prevptr,newptr,ptr)
        ; newptr^.par := abs(num)
        ; newptr^.val := sgn(num)
        end
      end
    end (* addpartolastrow *)

; procedure condeqn ( v1,v2,v3 : vector3 )
  (* Add the equation f(v1)+f(v2)+f(v3)=0 to the equations matrix,
     provided that the orbits of v2 and v3 vanish or do not precede
     that of v1. *)
  ; var
      on1,on2,on3,pr1,pr2,pr3 : integer
  ; begin
      locate(v1,on1,pr1)
    ; locate(v2,on2,pr2)
    ; locate(v3,on3,pr3)
    ; if ((pr2=0) or (on2>=on1)) and ((pr3=0) or (on3>=on1))
      then
      begin
        newrow
      ; addpartolastrow(pr1)
      ; addpartolastrow(pr2)
      ; addpartolastrow(pr3)
      ; remlastrowifzero
      end
    end (* condeqn *)

; procedure makeeqnmat
  (* Make a matrix with equations.  The equations are sorted by the
     lowest numbered orbit occurring in them which is not known to
     vanish. *)
  ; var
      i : integer
    ; v1,v2,v3 : vector3
  ; begin
      initeqnmatrix
    ; for i := 1 to orbits do
      with orbit[i] do
      if parref<>0 then (* otherwise the function vanishes *)
      begin
        makevecmodd(     x[1],     x[2],x[3],v1)
      ; makevecmodd(     x[2],x[1]-x[2],x[3],v2)
      ; makevecmodd(x[1]-x[2],     x[1],x[3],v3)
      ; condeqn(v1,v2,v3)
      ; makevecmodd(     x[1],     x[2],x[3],v1)
      ; makevecmodd(     x[2],x[1]+x[2],x[3],v2)
      ; makevecmodd(x[1]+x[2],     x[1],x[3],v3)
      ; condeqn(v1,v2,v3)
      ; makevecmodd(     x[2],     x[3],x[1],v1)
      ; makevecmodd(     x[3],x[2]-x[3],x[1],v2)
      ; makevecmodd(x[2]-x[3],     x[2],x[1],v3)
      ; condeqn(v1,v2,v3)
      ; makevecmodd(     x[2],     x[3],x[1],v1)
      ; makevecmodd(     x[3],x[2]+x[3],x[1],v2)
      ; makevecmodd(x[2]+x[3],     x[2],x[1],v3)
      ; condeqn(v1,v2,v3)
      ; makevecmodd(     x[3],     x[1],x[2],v1)
      ; makevecmodd(     x[1],x[3]-x[1],x[2],v2)
      ; makevecmodd(x[3]-x[1],     x[3],x[2],v3)
      ; condeqn(v1,v2,v3)
      ; makevecmodd(     x[3],     x[1],x[2],v1)
      ; makevecmodd(     x[1],x[3]+x[1],x[2],v2)
      ; makevecmodd(x[3]+x[1],     x[3],x[2],v3)
      ; condeqn(v1,v2,v3)
      end
    end (* makeeqnmat *)

(**********************************************************************)
(*                                                                    *)
(*                     Euclid--Gauss elimination                      *)
(*                                                                    *)
(**********************************************************************)

; function entry ( r,c : integer ) : integer
  (* Give entry (r,c) of the equations matrix. *)
  ; var
      ptr,prevptr : pointer
    ; found : boolean
  ; begin
      ptr := rowptr[r]
    ; prevptr := nil
    ; advanceto(prevptr,ptr,parofcol[c],found)
    ; if found
      then entry := ptr^.val
      else entry := 0
    end (* entry *)

; procedure getcol ( c : integer ; var cv : colvector )
  (* Copy a column of the equations matrix into an array. *)
  ; var
      i : integer
  ; begin
      for i := 1 to rows do cv[i] := entry(i,c)
    end (* getcol *)

; procedure minnon0incolvec ( var cv : colvector ; row1 : integer
            ; var rmin : integer ; var found : boolean )
  (* Look for a minimal non-zero number among cv[row1],...,cv[rows].  *)
  ; var
      min,i : integer
    ; foundone : boolean
  ; begin
      found := false
    ; if row1<=rows
      then
      begin
        foundone := false
      ; min := maxint
      ; i := row1
      ; repeat
          if cv[i]<>0 then
          begin
            found := true
          ; if abs(cv[i])<min then
            begin
              min := abs(cv[i])
            ; rmin := i
            ; foundone := (min=1)
            end
          end
        ; i := i+1
        until (i>rows) or foundone
      end
    end (* minnon0incolvec *)

; function gcdofcolvec ( cv : colvector ; row1 : integer ) : integer
  (* Calculate the g.c.d. of cv[row1],...,cv[rows] *)
  ; var
      rmin,rmin2,i : integer
    ; found : boolean
  ; begin
      minnon0incolvec(cv,row1,rmin2,found)
    ; if not found
      then gcdofcolvec := 0
      else
      begin
        repeat
          rmin := rmin2
        ; for i := row1   to rmin-1 do cv[i] := cv[i] mod cv[rmin]
        ; for i := rmin+1 to rows   do cv[i] := cv[i] mod cv[rmin]
        ; minnon0incolvec(cv,row1,rmin2,found)
        until (rmin=rmin2)
      ; gcdofcolvec := cv[rmin]
      end
    end (* gcdofcolvec *)

 ; procedure goodcol ( row1,col1 : integer ; var cgood : integer
      ; var found : boolean )
   (* Among the columns col1,...,params of the equations matrix. look
      for a column whose entries row1,...,rows have  minimal non-zero
      g.c.d. *)
  ; var
      absgcd,min,i : integer
    ; coli : colvector
    ; foundone : boolean
  ; begin
      found := false
    ; if (row1<=rows) and (col1<=params)
      then
      begin
        foundone := false
      ; min := maxint
      ; i := col1
      ; repeat
          getcol(i,coli)
        ; absgcd := abs(gcdofcolvec(coli,row1))
        ; if (absgcd<>0) and (absgcd<min)
          then
          begin
            found := true
          ; min := absgcd
          ; cgood := i
          ; foundone := (min=1)
          end
        ; i := i+1
        until (i>params) or foundone
      end
    end (* goodcol *)

; procedure setbadness ( r : integer )  (* not useful *)
  ; var
      ptr : pointer
  ; begin
      badness[r] := 0
    ; ptr := rowptr[r]
    ; while ptr<>nil do
      begin
        badness[r] := badness[r]+abs(ptr^.val)
      ; ptr := ptr^.next
      end
    end (* setbadness *)

; procedure goodrow ( row1 : integer ; var rgood : integer )
   (* Among the entries row1,...,rows of the good column thecol, look
      for an entry that is  minimal non-zero. The notion of badness is no 
      longer relevant. *)
  ; var
      r,minthecol: integer
    ; minbadn : real
  ; begin
      minthecol := maxint
    ; minbadn := maxint*maxint*maxint
    ; for r := row1 to rows do
      begin
        if thecol[r]<>0 then
        begin
          if abs(thecol[r])<minthecol
          then
          begin
            minthecol := abs(thecol[r])
          ; minbadn := badness[r]
          ; rgood := r
          end
          else
          if (abs(thecol[r])=minthecol) and (badness[r]<minbadn) then
          begin
            minbadn := badness[r]
          ; rgood := r
          end
        end
      end
    end (* goodrow *)

; function modbigprime ( a : integer ) : integer
  (* Reduce modulo bigprime in order to get it in the range between
     -bigprime/2 and +bigprime/2                                    *)
  ; begin
      a := a mod bigprime
    ; if a>bigprime div 2
      then modbigprime := a-bigprime
      else 
      if a<-bigprime div 2
      then modbigprime := a+bigprime
      else modbigprime := a
    end (* modbigprime *)

; procedure reduce ( r1,r2 : integer )
   (* subtract an appropriate multiple of the r1 row from the r2 row and 
      update the good column thecol. *)  
  ; var
      q : integer
    ; ptr1,ptr2,prevptr2,newptr2,nextptr2 : pointer
    ; found : boolean
  ; begin
      q := thecol[r2] div thecol[r1]
    ; if q<>0 then
      begin
        ptr1 := rowptr[r1]
      ; ptr2 := rowptr[r2]
      ; prevptr2 := nil
      ; while ptr1<>nil do
        begin
          advanceto(prevptr2,ptr2,ptr1^.par,found)
        ; if found then
          begin
            ptr2^.val := modbigprime(ptr2^.val-q*ptr1^.val)
          ; if ptr2^.val=0
            then
            begin
              nextptr2 := ptr2^.next
            ; delete(rowptr[r2],prevptr2,ptr2)
            ; ptr2 := nextptr2
            end
          end
          else
          begin
            insert(rowptr[r2],prevptr2,newptr2,ptr2)
          ; prevptr2 := newptr2
          ; newptr2^.par := ptr1^.par
          ; newptr2^.val := modbigprime(-q*ptr1^.val)
          end
        ; ptr1 := ptr1^.next
        end
      ; thecol[r2] := modbigprime(thecol[r2]-q*thecol[r1])
      ; setbadness(r2)
      end
    end (* reduce *)

; procedure swaprows ( r1,r2 : integer )
  (* Swap two rows of the equations matrix. *)
  ; var
      h : pointer
  ; begin
      h := rowptr[r1]
    ; rowptr[r1] := rowptr[r2]
    ; rowptr[r2] := h
    end (* swaprows *)

; procedure swapcols ( c1,c2 : integer )
  (* Swap two columns of the equations matrix. *)
  ; var
      h : real
  ; begin
      swapints(parofcol[c1],parofcol[c2])
    ; colofpar[parofcol[c1]] := c1
    ; colofpar[parofcol[c2]] := c2
    ; swapints(thecol[c1],thecol[c2])
    ; h := badness[c1]
    ; badness[c1] := badness[c2]
    ; badness[c2] := h
    end (* swapcols *)

; procedure remzerorows ( row1 : integer )
  ; var
      i : integer
  ; begin
      i := row1
    ; while i<=rows do
      if rowptr[i]<>nil
      then i := i+1
      else
      begin
        rowptr[i] := rowptr[rows]
      ; rows := rows-1
      end
    end (* remzerorows *)

; procedure euclidgauss
  (* Transforms the equations matrix into diagonal form, followed by "dim"
     columns. Zero rows are discarded. *)  
  ; var
      rgood,rgood2,cgood,i : integer
    ; found : boolean
  ; begin
      diag := 0
    ; for i := 1 to rows do setbadness(i)
    ; repeat
        goodcol(diag+1,diag+1,cgood,found)
      ; if found then
        begin
          getcol(cgood,thecol)
        ; goodrow(diag+1,rgood2)
        ; repeat
            rgood := rgood2
          ; for i :=       1 to rgood-1 do reduce(rgood,i)
          ; for i := rgood+1 to rows    do reduce(rgood,i)
          ; goodrow(diag+1,rgood2)
          until rgood2=rgood
        ; diag := diag+1
        ; swaprows(rgood,diag)
        ; swapcols(cgood,diag)
        ; remzerorows(diag+1)
        end
      until not found
    end (* euclidgauss *)

; procedure calcsolnspace
  (* Calculate a basis of the solution space from the equations matrix
     after Euclid--Gauss elimination has been carried out. 
     The parofcol encodes the column swaps that have been made; 
     if no swaps have been made, then parofcol[i]=i. 
     If s<=dim, t<=dim then the entry soln[s,freepar[t]] becomes non-zero 
     iff s=t.                                                          *)
  ; var
      s,i,j,sum,factor : integer
  ; begin
      dim := params-diag
    ; for s := 1 to dim do
      begin
        freepar[s] := parofcol[diag+s]
      ; for i := diag+1 to params do soln[s,parofcol[i]] := 0
      ; soln[s,parofcol[diag+s]] := 1
      ; for i := diag downto 1 do
        begin
          sum := 0
        ; for j := i+1 to params do
          sum := modbigprime(sum+entry(i,j)*soln[s,parofcol[j]])
        ; factor := abs(entry(i,i) div gcd2(entry(i,i),sum))
        ; soln[s,parofcol[i]] := modbigprime(-(factor*sum) div entry(i,i))
        ; for j := i+1 to params do
          soln[s,parofcol[j]] := modbigprime(factor*soln[s,parofcol[j]])
        end
      end
   end (* calcsolnspace *)

(**********************************************************************)
(*                                                                    *)
(*                       input/output routines                        *)
(*                                                                    *)
(**********************************************************************)
 
; procedure writesolnspace
  ; var
      i,s : integer
  ; begin
      writeln(d:1)
    ; writeln(params:1)
    ; writeln(dim:1)
    ; for s := 1 to dim do
      begin
        write(freepar[s]:1)
      ; if (s mod 10=0) or (s=dim)
        then writeln
        else write(tab)
      end
    ; for s := 1 to dim do
      for i := 1 to params do
      begin
        write(soln[s,i]:1)
      ; if (i mod 10=0) or (i=params)
        then writeln
        else write(tab)
      end
    end

; begin
    readorbits
  ; make2tables
  ; makeparams
  ; makeeqnmat
  ; euclidgauss
  ; calcsolnspace
  ; writesolnspace
  end.
